--- /dev/null	2013-02-24 04:11:47.000000000 +0000
+++ new/usr/src/cmd/iconv/charmap.c	2013-02-24 04:11:47.644924595 +0000
@@ -0,0 +1,147 @@
+/*
+ * This file and its contents are supplied under the terms of the
+ * Common Development and Distribution License ("CDDL"), version 1.0.
+ * You may only use this file in accordance with the terms of version
+ * 1.0 of the CDDL.
+ *
+ * A full copy of the text of the CDDL should have accompanied this
+ * source.  A copy of the CDDL is also available via the Internet at
+ * http://www.illumos.org/license/CDDL.
+ */
+
+/*
+ * Copyright 2010 Nexenta Systems, Inc.  All rights reserved.
+ * Copyright 2013 David Hoeppner.  All rights reserved.
+ */
+
+/*
+ * Character map handling for iconv.
+ */
+
+#include <sys/avl.h>
+
+#include <stddef.h>
+
+#include "iconv.h"
+#include "parser.tab.h"
+
+/*
+ * AVL trees for the from and to charmaps.
+ */
+static avl_tree_t	from_cmap_sym;
+static avl_tree_t	from_cmap_wc;
+static avl_tree_t	to_cmap_sym;
+static avl_tree_t	to_cmap_wc;
+
+static avl_tree_t	*current_cmap_sym = &from_cmap_sym;
+static avl_tree_t	*current_cmap_wc = &from_cmap_wc;
+
+typedef struct charmap {
+	const char	*name;
+	wchar_t		wc;
+	avl_node_t	avl_sym;
+	avl_node_t	avl_wc;
+} charmap_t;
+
+static int
+cmap_compare_sym(const void *n1, const void *n2)
+{
+	const charmap_t	*c1 = n1;
+	const charmap_t	*c2 = n2;
+	int	rv;
+
+	rv = strcmp(c1->name, c2->name);
+	return ((rv < 0) ? -1 : (rv > 0) ? 1 : 0);
+}
+
+static int
+cmap_compare_wc(const void *n1, const void *n2)
+{
+	const charmap_t	*c1 = n1;
+	const charmap_t	*c2 = n2;
+
+	return ((c1->wc < c2->wc) ? -1 : (c1->wc > c2->wc) ? 1 : 0);
+}
+
+void
+init_charmap(void)
+{
+	avl_create(&from_cmap_sym, cmap_compare_sym, sizeof (charmap_t),
+	    offsetof(charmap_t, avl_sym));
+
+	avl_create(&from_cmap_wc, cmap_compare_wc, sizeof (charmap_t),
+	    offsetof(charmap_t, avl_wc));
+
+	avl_create(&to_cmap_sym, cmap_compare_sym, sizeof (charmap_t),
+	    offsetof(charmap_t, avl_sym));
+
+	avl_create(&to_cmap_wc, cmap_compare_wc, sizeof (charmap_t),
+	    offsetof(charmap_t, avl_wc));
+}
+
+/*
+ * Switches from fromcharmap to tocharmap.
+ */
+void
+switch_charmap(void)
+{
+	current_cmap_sym = &to_cmap_sym;
+	current_cmap_wc = &to_cmap_wc;
+}
+
+static void
+add_charmap_impl(char *sym, wchar_t wc, int nodups)
+{
+	charmap_t	srch;
+	charmap_t	*n = NULL;
+	avl_index_t	where;
+
+	srch.wc = wc;
+	srch.name = sym;
+
+	/*
+	 * Also possibly insert the wide mapping, although note that there
+	 * can only be one of these per wide character code.
+	 */
+	if ((wc != -1) && ((avl_find(current_cmap_wc, &srch, &where)) == NULL)) {
+		if ((n = calloc(1, sizeof (*n))) == NULL) {
+			errf(_("out of memory"));
+			return;
+		}
+
+		n->wc = wc;
+		avl_insert(current_cmap_wc, n, where);
+	}
+
+	if (sym != NULL) {
+		if (avl_find(current_cmap_sym, &srch, &where) != NULL) {
+			if (nodups == 1) {
+				errf(_("duplicate character definition"));
+			}
+
+			return;
+		}
+
+		if ((n == NULL) && ((n = calloc(1, sizeof (*n))) == NULL)) {
+			errf(_("out of memory"));
+			return;
+		}
+
+		n->wc = wc;
+		n->name = sym;
+printf("ADDING %s\n", sym);
+		avl_insert(current_cmap_sym, n, where);
+	}
+}
+
+void
+add_charmap(char *sym, int c)
+{
+	add_charmap_impl(sym, c, 1);
+}
+
+void
+add_charmap_range(char *s, char *e, int wc)
+{
+
+}
